const crypto = require('crypto');
const formidable = require('formidable');
const zlib = require('zlib');
const sqlite3 = require('sqlite3').verbose();
const https = require('https');
const nodemailer = require('nodemailer');


module_core_email_alert = function(system, info, callback=(err, info)=>{}) {
    if(SYSTEM.mailServer[system.sysId]) {
        let options = {};
        if(SYSTEM.mailServer[system.sysId].default) {
            options = {
                service:'gmail', 
                auth:SYSTEM.mail
            };    
        } else {
            options = {
                host:SYSTEM.mailServer[system.sysId].smtp,
                port:+SYSTEM.mailServer[system.sysId].port,
                secure:+SYSTEM.mailServer[system.sysId].port==465,
                auth:{
                    user:SYSTEM.mailServer[system.sysId].user, 
                    pass:SYSTEM.mailServer[system.sysId].pass
                }
            }
        }
        const transport = nodemailer.createTransport(options);
        transport.sendMail({
            from: SYSTEM.mailServer[system.sysId].default?"no-reply@aimagin.com":options.auth.user,
            to: info.to,
            cc: info.cc||"",
            bcc: info.bcc||"",
            subject: info.subject,
            text:info.text,
            html: info.html,
        }, callback);          
    } else {
        return null;
    }
};

module_core_email = function(info, callback=(e)=>{}) {
    if(CUSTOM_MODULE.module_custom_email) {
        CUSTOM_MODULE.module_custom_email(info, callback);
    } else {
        let sendData = JSON.stringify({data:info});
        let req = https.request({
            hostname: 'iot.aimagin.com',
            path: '/T2WnSQaNNdmbn5g24zGe4tvwm4UTvTsR',
            method: 'POST',
            headers: {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Content-Length': Buffer.byteLength(sendData)
            }
        }, (res) => {
            res.setEncoding('utf8');
            let response = "";
            res.on('data', (chunk) => {
                response+=chunk;
            }).on('end', () => {
                callback(JSON.parse(response));
    //            console.log(response);  //response.error, response.info.envelope, response.info.response
            });                            
        }).on('error', (e) => {
            callback(e)
        });
        req.end(sendData);        
    }
}

module_core_getTimeMsStr = function() {
    return new Date().getTime()+"";
}

module_core_parseCookie = function(cookie) {
    let result = {};
    if(typeof cookie == "string") {
        cookie.split(";").forEach(function(d) {
            d = d.trim().split("=");
            result[d[0]] = d[1];
        });
    }
    return result;
}

module_core_parsePost = function(system, callback) {
    system.form = new formidable.IncomingForm({
        maxFileSize:10*1024*1024*1024,
        uploadDir:system.sysPath+"/drive/upload",
        multiples:true,
        hash:'md5'
    }).parse(system.req, callback);
}

module_core_hash = function(key) {
    let md5sum = crypto.createHash('md5');
    md5sum.update(key);
    return md5sum.digest('hex');
}

module_core_random = function(n=64) {
    let str = ["ABCDEFGHIJKLMNOPQRSTUVWXYZ", "abcdefghijklmnopqrstuvwxyz", "0123456789"+new Date().getTime()];
    let code = [];
    for(let i=0;i<n;i++) {
        let r0 = Math.round(1+Math.random()*(str.length-1))-1;
        let r1 = Math.round(1+Math.random()*(str[r0].length-1))-1;
        code.push(str[r0][r1]);
    }
    for(let i=0;i<n;i++) {
        let t0 = Math.round(1+Math.random()*(code.length-1))-1;
        let t1 = Math.round(1+Math.random()*(code.length-1))-1;
        let temp = code[t0];
        code[t0] = code[t1];
        code[t1] = temp;
    }
    return code.join("");
};

module_core_genId = function(n=64) {
    return "_"+new Date().getTime()+"_"+module_core_random(n);
}

module_core_genVarName = function(n=64) {
    let c = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
    return c[Math.round(Math.random()*(c.length-1))]+module_core_random(n);
}


module_core_genDbIdVarName = function(n=64) {
    return new Date().toISOString().replace(/-|T|:|\.|Z/g, '')+module_core_random(n);
}

module_core_genDbId = function(n=16) {
    return new Date().toISOString().replace(/-|T|:|\.|Z/g, '')+module_core_random(n);
}

module_core_gzip = function(src, dst, callback) {
    const gzip = zlib.createGzip();
    fs.createReadStream(src).pipe(gzip)
      .on('error', () => {
        // handle error
      })
      .pipe(fs.createWriteStream(dst))
      .on('error', () => {
        // handle error
      }).on("finish", ()=>{
        callback();
      });
}

module_core_checkPermission = function(system, id, mode, callback) {
    if(mode=="module") {
        let mod = id.split("-")[0];
        callback(system.conf.module[mod]!=undefined && system.conf.module[mod].status==true && (system.role.findIndex(d=>d.id=="admin")>=0 || system.role.some(d=>d[mode].indexOf(id)>=0)));
    } else {
        callback((system.role.findIndex(d=>d.id=="admin")>=0 && (mode=="database" || mode=="page")) || system.role.some(d=>d[mode].indexOf(id)>=0));
    }
}

module_core_selectOnlySql = function(sql) {
    return sql.replace(/(DROP|CREATE|ALTER)( )*TABLE|DELETE|INSERT/gi, '');
}

module_core_recordRequestLog = function(system, userEmail, mode, config){
    const now = new Date();
    const milli = Date.now();
    const y = now.getFullYear();
    const m = now.getMonth()+1;
    const m_str = m<10 ? "0"+m.toString() : m;
    const d = now.getDate();
    const h = now.getHours();
    const mi = now.getMinutes();
    const s = now.getSeconds();
    const mil = now.getMilliseconds();
    const d_str = d<10 ? "0"+d.toString() : d;
    const h_str = h<10 ? "0"+h.toString() : h;
    const mi_str = mi<10 ? "0"+mi.toString() : mi;
    const s_str = s<10 ? "0"+s.toString() : s;
    const sysId = system.cookie._sysId;

    let mil_str="";
    if (mil >= 100){
        mil_str = mil;
    }
    else if (mil>= 10) {
        mil_str = "0"+mil;
    }
    else{
        mil_str = "00"+mil;
    }

    const local = y+"-"+m_str+"-"+d_str+" "+h_str+":"+mi_str+":"+s_str+"."+mil_str;
  
    const removeOldDatabase = function(path) {
        let files = fs.readdirSync(path);
        files = files.sort(function(a, b){
            let a_split = a.replace('.db','').split('_');
            let b_split = b.replace('.db','').split('_');
            return Date.UTC(a_split[0], a_split[1],a_split[2],a_split[3]) - Date.UTC(b_split[0], b_split[1],b_split[2],b_split[3])
            });
        if (files.length > 725){
            fs.unlinkSync(path+"/"+files[0]);
        }
    }

    if(mode == "page"){
        if(system.req.method == "GET"){
            let mode_path="./system_log/user_request/";

            //Preparing folder
            if(!fs.existsSync("./system_log/")){
                fs.mkdirSync("./system_log/");
            }
            if(!fs.existsSync(mode_path)){
                fs.mkdirSync(mode_path);
            }
            if(!fs.existsSync(mode_path+sysId)){
                fs.mkdirSync(mode_path+sysId);
            } 

            //Databse writing
            let db = new sqlite3.Database(mode_path+sysId+"/"+y+"_"+m+"_"+d+"_"+h+".db"); 
            db.serialize(()=>{
                db.run('CREATE TABLE IF NOT EXISTS UserRequest_LOG (UserId TEXT, Email TEXT, Referer TEXT, Mobile TEXT, Platform TEXT, SEC_CH_UA TEXT, UrlRequest TEXT, TimeStamp TEXT, Time REAL)');

                db.run('INSERT INTO UserRequest_LOG (UserId, Email, Referer, Mobile, Platform, SEC_CH_UA, UrlRequest, TimeStamp, Time)  VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)', 
                [   
                    system.cookie._uId,
                    userEmail,
                    system.req.headers.referer,
                    system.req.headers['sec-ch-ua-mobile'],
                    system.req.headers['sec-ch-ua-platform'],
                    system.req.headers['sec-ch-ua'],
                    system.req.url,
                    local,
                    milli,
                ], function(err) {
                    if (err) {
                        fs.appendFileSync("./process/module_core_log.log", "9988|-----module_core_recordRequestLog-----|"+"\r\n"+err + new Date().toISOString()+"\r\n");                        
                        db.close(); 
                    }
                    else{   
                        db.close();    
                    }
                });

            });

            removeOldDatabase(mode_path+sysId);
        }
    }
    else if(mode == "sqlite"){
        if(system.req.method == "POST"){
            let mode_path="./system_log/sqlite_request/";

            //Preparing folder
            if(!fs.existsSync("./system_log/")){
                fs.mkdirSync("./system_log/");
            }
            if(!fs.existsSync(mode_path)){
                fs.mkdirSync(mode_path);
            }
            if(!fs.existsSync(mode_path+sysId)){
                fs.mkdirSync(mode_path+sysId);
            } 

            //Databse writing
            let db = new sqlite3.Database(mode_path+sysId+"/"+y+"_"+m+"_"+d+"_"+h+".db"); 
            db.serialize(()=>{
                db.run('CREATE TABLE IF NOT EXISTS SqliteRequest_LOG (UserId TEXT, Email TEXT, Referer TEXT, Mobile TEXT, Platform TEXT, SEC_CH_UA TEXT, UrlRequest TEXT, TimeStamp TEXT, Time REAL, ProcessedTime REAL, APIName TEXT, RequestConfig TEXT, ErrorMsg TEXT)');

                db.run('INSERT INTO SqliteRequest_LOG (UserId, Email, Referer, Mobile, Platform, SEC_CH_UA, UrlRequest, TimeStamp, Time, ProcessedTime, APIName, RequestConfig, ErrorMsg)  VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)', 
                [   
                    system.cookie._uId,
                    userEmail,
                    system.req.headers.referer,
                    system.req.headers['sec-ch-ua-mobile'],
                    system.req.headers['sec-ch-ua-platform'],
                    system.req.headers['sec-ch-ua'],
                    system.req.url,
                    local,
                    milli,
                    config.processTime,
                    config.apiName,
                    config.config,
                    config.errMsg
                ], function(err) {
                    if (err) {
                        fs.appendFileSync("./process/module_core_log.log", "9988|-----module_core_recordSQLiteRequestLog-----|"+"\r\n"+err + new Date().toISOString()+"\r\n");                        
                        db.close(); 
                    }
                    else{   
                        db.close();    
                    }
                });

            });

            removeOldDatabase(mode_path+sysId);
        }
    }
}