module_esp32 = function(mode) {
    let fn = null;
    switch(mode) {
        case "clientActive": fn = module_esp32_clientActive; break;
        case "clientInactive": fn = module_esp32_clientInactive; break;
        case "getFileList": fn = module_esp32_getFileList; break; 
        case "downloadFile": fn = module_esp32_downloadFile; break;
        case "downloadCSV": fn = module_esp32_downloadCSV; break;
        case "getSSIDs": fn = module_esp32_getSSIDs; break;
        case "mode1": fn = module_esp32_mode1; break;
        case "mode2": fn = module_esp32_mode2; break;
        case "get_sdcard_informations": fn = module_esp32_test; break;
        case "delete_logged_files": fn = module_esp32_delete_logged_files; break;
        case "set_time": fn = module_esp32_test; break;
        case "upload_certificates": fn = module_esp32_test; break;
        case "get_raw_values": fn = module_esp32_get_raw_values; break;
        case "get_battery_information": fn = module_esp32_get_battery_information; break;        
    }
    return fn;
}

function module_esp32_get_battery_information(system, data, callback) {
    callback({status:true, res:'success',  data:{remaining_batt:50, remaining_charge_time:10, batt_volt:5}});
}


function module_esp32_delete_logged_files(system, data, callback) {
    if(data.reset===true) {
        data.path = "/sdcard/logged_data/";
        if(fs.existsSync("./images_test/sdcard/logged_data")) {
            fs.readdirSync("./images_test/dummies/").forEach(function(d) {
                fs.copyFileSync("./images_test/dummies/"+d, "./images_test/sdcard/logged_data/"+d)
            });
            callback({status:true, res:'success', data:fs.readdirSync("./images_test/sdcard/logged_data")});
        } else {
            callback({status:true, res:'Folder not found.', data:[]});
        }
        return;
    }
    if(fs.existsSync("./images_test/sdcard/logged_data")) {
        fs.readdirSync("./images_test/sdcard/logged_data").forEach(function(d) {
            fs.unlinkSync("./images_test/sdcard/logged_data/"+d);
        });
        callback({status:true, res:'success', data:fs.readdirSync("./images_test/sdcard/logged_data")});
    } else {
        callback({status:true, res:'Folder not found.', data:[]});
    }
}

function module_esp32_get_raw_values(system, data, callback) {
    if(+data.len) {
        let d = [];
        for(let i=0;i<+data.len;i++) {
            d.push(Math.random());
        }
        callback({status:true, data:d});
    } else {
        callback({status:true, data:[]});
    }
}

function module_esp32_test(system, data, callback) {
    callback({status:true, res:'success', data});
}

function module_esp32_clientActive(system, data, callback) {
    console.log("active");
    callback({data:'return from clientActive'});
}

function module_esp32_clientInactive(system, data, callback) {
    console.log("Inactive");
    callback({data:'return from clientInactive'});
}

function module_esp32_getFileList(system, data, callback) {
    if(data.path) {
        if(fs.existsSync("./images_test/"+data.path)) {
            callback({data:fs.readdirSync("./images_test/"+data.path)});
        } else {
            callback({data:[]});
        }
    } else if(fs.existsSync("./images_test")) {
        callback({data:fs.readdirSync("./images_test")});
    } else {
        callback({data:[]});
    }
}

function module_esp32_getSSIDs(system, data, callback) {
    callback({data:["SSID-1", "SSID-2"]});
}

function module_esp32_downloadFile(system, data, callback) {
    if(data.db=="1") {
        system.res.writeHead(200, {'Content-disposition': 'attachment; filename='+encodeURI("database.db")});              
        fs.createReadStream("./systems/system1/db/database.db").pipe(system.res);
    } else if(fs.existsSync("./images_test/"+data.fileName)) {
        system.res.writeHead(200, {'Content-disposition': 'attachment; filename='+encodeURI(data.fileName)});              
        fs.createReadStream("./images_test/"+data.fileName).pipe(system.res);
    } else {
        system.res.end(null);
    }
}

function module_esp32_downloadCSV(system, data, callback) {
    console.log(data);
    system.res.writeHead(200, {'Content-disposition': 'attachment; filename='+encodeURI(data.filename)});              
    fs.createReadStream("./images_test/test.csv").pipe(system.res);
}

function module_esp32_mode1(system, data, callback) {
    callback({data:["this is mode1", data, "received"]});
}

function module_esp32_mode2(system, data, callback) {
    callback({data:["this is mode2", data, "received"]});
}