const http = require('http');
const url = require('url');
const sqlite3 = require('sqlite3').verbose();
const fs = require('fs');
const { ClickHouse } = require('clickhouse');

const SYSTEM = {
    hostname: null,
    //    path:require('path').dirname(require.main.filename),
    module: {}
};

(function () {
    const system = JSON.parse(fs.readFileSync("./core/config/system.json"));
    module.exports.port = system.port;
    SYSTEM.serverType = system.serverType;
    SYSTEM.origin = system.origin;
    SYSTEM.sysId = system.sysId;
    SYSTEM.oId = system.oId;
    SYSTEM.path = system.path;
    SYSTEM.sysPath = SYSTEM.path + "systems/" + system.sysId;
    SYSTEM.logInText = system.log_in_text;
    SYSTEM.clickhousepwd = system.clickhousepwd;
    SYSTEM.clickhouse = system.clickhouse;
    SYSTEM.ext_clickhouse = system.ext_clickhouse;
    SYSTEM.mariadb = system.mariadb;    
    SYSTEM.mqtt = system.mqtt;
    SYSTEM.mail = system.mail;
    SYSTEM.backup = {...system.backup,web_config: undefined, current_token: undefined};
    console.log("SYSTEM: " + JSON.stringify(SYSTEM));
})();

/* included files*/
const CUSTOM_MODULE = {};
if (fs.existsSync("./core/custom/module_custom.js")) {
    eval(fs.readFileSync("./core/custom/module_custom.js", "utf8"));
}
const config = JSON.parse(fs.readFileSync("./core/config/config.json"));
config.module.forEach(function (d) {
    eval(fs.readFileSync("./core/server/" + d.file, "utf8"));
});
/* included files*/

function module_main_responseGzip(res, path, headers) {
    headers['Content-Encoding'] = "gzip";
    res.writeHead(200, headers);
    if (!fs.existsSync(path)) {
        path = "./core/client/err_404.html.gz";
    }
    fs.createReadStream(path).pipe(res);
}

function module_main_responseJson(res, status, data, cookie, statusCode) {
    let headers = {
        "Content-Type": "application/json"
    };
    if (cookie != null && cookie != undefined) headers["Set-Cookie"] = cookie;
    res.writeHead(statusCode ? statusCode : 200, headers);
    res.end(JSON.stringify({
        res: status,
        data
    }));
}

function module_main_include(qpath, res) {
    let path = SYSTEM.path + qpath.join("/") + ".gz";
    if (fs.existsSync(path)) {
        let fileType = qpath[qpath.length - 1].split(".").pop();
        let type = "text/plain";
        if (fileType == "css") {
            type = 'text/css';
        } else if (fileType == "js") {
            type = 'text/javascript';
        } else if (fileType == "ttf" || fileType == "woff2" || fileType == "woff" || fileType == "eot") {
            type = 'font/' + fileType;
        } else if (fileType == "png" || fileType == "jpeg" || fileType == "jpg") {
            type = 'image/' + fileType;
        }
        module_main_responseGzip(res, path, {
            "Content-Type": type,
            "Cache-Control": "public, max-age=86400, stale-while-revalidate=3600;"
        });
    } else if (qpath[0] == "favicon.ico") {
        module_main_responseGzip(res, "./include/img/waijung_favicon.ico.gz", {
            "Content-Type": "image/x-icon",
            "Cache-Control": "public, max-age=864000, stale-while-revalidate=3600;"
        });
    } else {
        fs.appendFileSync("include_not_found.log", path + "\r\n"); /*for server only*/
        res.end(null);
    }
}

function module_main_getInitInfo(system) {
    let webAppMenu, webApp, c = 2;
    console.log("getInitInfo: " + system.sysPath + "/db/system.db");
    let db = new sqlite3.Database(system.sysPath + "/db/system.db");
    db.parallelize(() => {
        db.all("SELECT id, name, icon FROM tbl005 WHERE menu = '1'", [], (err, rows) => {
            webApp = rows;
            next();
        });
        db.get("SELECT data FROM tbl006 WHERE id = ?", ["webAppMenu"], (err, row) => {
            webAppMenu = row.data;
            next();
        });

        function next() {
            if (--c > 0) return;
            module_main_responseJson(system.res, true, {
                user: system.user,
                role: system.role,
                webApp: webApp,
                webAppMenu: webAppMenu,
                oId: system.oId,
                conf: JSON.parse(fs.readFileSync(system.sysPath + "/conf/conf.json"))
            });
        }
    });
    db.close();
}

function module_main_getAppIdByPath(system, callback) {
    let db = new sqlite3.Database(system.sysPath + "/db/system.db");
    db.get("SELECT id FROM tbl005 WHERE path = ?", [system.publicPath], (err, webApp) => {
        callback(webApp == undefined ? null : webApp.id);
    });
    db.close();
}

function module_main_errJson(err) {
    switch (err) {
        case "err_1_1":
            module_main_responseJson(system.res, false, {
                ref: "err_0_1"
            });
            break; /*inactive user*/
        case "err_1_2":
            module_main_responseJson(system.res, false, {
                ref: "err_0_2"
            });
            break; /*no_role_user*/
        case "err_1_3":
            module_main_responseJson(system.res, false, {
                ref: "err_0_3"
            });
            break; /*incorrect sessionId*/
        case "err_1_4":
            module_main_responseJson(system.res, false, {
                ref: "err_0_4"
            });
            break; /*incorrect oId*/
    }
}

function module_main_errPage(system, cookie, ref) {
    let file = null;
    switch (ref) {
        case "err_0_0":
            file = "err_login";
            break;
        case "err_0_1":
        case "err_1_1":
            file = "err_inactive";
            break;
        case "err_0_2":
        case "err_1_2":
            file = "err_role";
            break;
        case "err_0_3":
        case "err_1_3":
            file = "err_session";
            break;
        case "err_0_4":
        case "err_1_4":
            file = "err_internal";
            break;
        case "err_0_5":
            file = "err_404";
            break;
        default:
            file = "logout";
    }
    module_main_responseGzip(system.res, "./core/client/" + file + ".html.gz", {
        'Content-Type': "text/html",
        "Set-Cookie": cookie
    });
}

function module_main_page(system, cookie) {
    if (system.qdata._pageId != undefined) {
        module_core_checkPermission(system, system.qdata._pageId, "page", (result) => {
            module_main_responseGzip(system.res, result ? "./core/client/" + system.qdata._pageId + ".html.gz" : "./core/client/err_404.html.gz", {
                "Content-Type": "text/html",
                "Set-Cookie": cookie
            });
        });
    } else if (system.qdata._appId != undefined) {
        module_core_checkPermission(system, system.qdata._appId, "webApp", (result) => {
            module_main_responseGzip(system.res, result ? system.sysPath + "/webApp/" + system.qdata._appId + ".html.gz" : "./core/client/err_404.html.gz", {
                "Content-Type": "text/html",
                "Set-Cookie": cookie
            });
        });
    } else if (system.qdata._previewId != undefined) {
        module_main_responseGzip(system.res, system.sysPath + "/preview/" + system.qdata._previewId + ".html.gz", {
            "Content-Type": "text/html",
            "Set-Cookie": cookie
        });
    } else {
        module_main_responseGzip(system.res, "./core/client/login_success.html.gz", {
            "Content-Type": "text/html",
            "Set-Cookie": cookie
        });
    }
}

function module_main_get(system) {
    if (system.qpath[0] == 'esp32-downloadFile') {
        module_auth_authenticate(system, (status, err, cookie) => {
            module_esp32("downloadFile")(system, {
                fileName: system.qdata.file,
                db: system.qdata.db
            }, (data) => { });
        });
        return;
    } else if (system.qpath[0] == 'esp32-downloadCSV') {
        module_auth_authenticate(system, (status, err, cookie) => {
            module_esp32("downloadCSV")(system, {
                filename: decodeURI(system.qdata.file),
                sql: decodeURI(system.qdata.sql)
            }, (data) => { });
        });
        return;
    
    }
    if (system.qdata._apiId != undefined) {
        module_main_responseGzip(system.res, "./core/client/" + system.qdata._apiId + ".js.gz", { 'Content-Type': "text/html" });
    } else if (system.qdata._ == "logout") {
        module_auth_logOut(system, (cookie) => module_main_errPage(system, cookie, system.qdata._ref));
    } else if (system.qdata._pageId != undefined || system.qdata._appId != undefined) {
        module_auth_authenticate(system, (status, err, cookie) => {
            cookie.push('AIMAGIN_ACCESS_TYPE=private;');
            status ? module_main_page(system, cookie) : module_main_errPage(system, cookie, err);
        });
    } else if (system.qdata._exportId != undefined) {
        /*server only*/
        module_auth_authenticate(system, (status, err, cookie) => {
            if (status) {
                module_server("4")(system, {
                    id: system.qdata._exportId,
                    name: system.qdata._exportName
                }, () => {
                    module_main_errPage(system, cookie, "err_0_5");
                });
            } else {
                module_main_responseJson(system.res, false, {
                    ref: err
                }, cookie);
            }
        });
    } else if (system.qdata._fileId != undefined) {
        module_auth_authenticate(system, (status, err, cookie) => {
            //if(status) {
            if (system.user == undefined) {
                system.user = { userId: "Public" };
            }
            module_server("9")(system, {
                id: system.qdata._fileId,
                download: system.qdata._download
            }, (data) => {
                //file not found
                module_main_responseGzip(system.res, "./core/client/err_404.html.gz", {
                    "Content-Type": "text/html",
                    "Set-Cookie": cookie
                });
                //module_main_errPage(system, cookie, "err_0_5");
            });
            //} else {
            //    module_main_responseJson(system.res, false, {ref:err}, cookie);
            //}
        });
    } else if (system.qdata._csvId != undefined) {
        /*server only*/
        module_auth_authenticate(system, (status, err, cookie) => {
            if (status) {
                module_python(system, 'downloadCSV')(system, {
                    id: system.qdata._csvId
                }, (data) => {

                });
            } else {
                module_main_responseJson(system.res, false, {
                    ref: err
                }, cookie);
            }
        });
    } else if (system.qdata._previewId != undefined) {
        /*server only*/
        module_auth_authenticate(system, (status, err, cookie) => {
            cookie.push('AIMAGIN_ACCESS_TYPE=preview;');
            status ? module_main_page(system, cookie) : module_main_errPage(system, cookie, err);
        });
    } else if (system.qdata._file_backup != undefined) {
        module_auth_authenticate(system, (status, err, cookie) => {
            try{
                // if (!status || system.user == undefined || (system?.user?.roleId && !system?.user?.roleId.split("|").includes("admin"))) throw "err_0_2";

                if (!system.qdata._file_backup.includes("-")) throw "err_0_5"

                const func = module_backup("download_file");

                let file_obj = system.qdata._file_backup.split("-");

                if (!func) throw "err_0_4"

                func(system, {process: file_obj[0], file: file_obj[1]}, (data) => {
                    module_main_errPage(system,cookie,"err_0_5")
                });
            }catch(e){
                module_main_errPage(system,cookie,e)
            }
        });
    
    } else if (system.qdata._backup_realtime != undefined){
        module_auth_authenticate(system, (status, err, cookie) => {
            try{
                // if (!status || system.user == undefined || (system?.user?.roleId && !system?.user?.roleId.split("|").includes("admin"))) throw "err_0_2";

                const func = module_backup("backup_log_realtime");
                if(!func) throw "err_0_4";
                func(system, {type: system.qdata._backup_realtime});

            }catch(e){
                module_main_errPage(system,cookie,e);
            }
        });
    } else if (system.public == true) {
        module_main_getAppIdByPath(system, function (appId) {
            if (appId == null) {
                module_main_responseGzip(system.res, "./core/client/err_404.html.gz", {
                    "Content-Type": "text/html",
                    "Set-Cookie": system.cookie0
                });
            } else {
                moudule_auth_genPublicCookie(system, (cookie) => module_main_responseGzip(system.res, system.sysPath + "/webApp/" + appId + ".html.gz", {
                    "Content-Type": "text/html",
                    "Set-Cookie": cookie
                }));
            }
        });   
    } else {
        module_auth_authenticate(system, (status, err, cookie) => {
            if (system.defaultPage == true) {
                if (status) {
                    if (system.oIdInPath == true) {
                        module_main_responseGzip(system.res, "./core/client/redirect.html.gz", {
                            "Content-Type": "text/html"
                        });
                    } else {
                        cookie.push('AIMAGIN_ACCESS_TYPE=private;');
                        module_main_page(system, cookie);
                    }
                } else {
                    module_main_responseGzip(system.res, "./core/client/login.html.gz", {
                        "Content-Type": "text/html",
                        "Set-Cookie": cookie
                    });
                }
            } else {
                module_main_responseGzip(system.res, "./core/client/err_404.html.gz", {
                    "Content-Type": "text/html",
                    "Set-Cookie": cookie
                });
            }
        });
    }
}

function module_main_genResetPasswordLink(sysId, data) {
    let db = new sqlite3.Database(SYSTEM.path + "systems/" + sysId + "/db/system.db", sqlite3.OPEN_READONLY);
    db.get("SELECT userId FROM tbl000 WHERE lower(email)=?", [data.email.toLowerCase()], (err, row) => {
        if (err == null && row !== undefined) {
            let id = module_core_genId(128);
            let link = SYSTEM.origin + "/yzbtpnaXb2NKC8YTq8naEqL4TcbLBNCQ94VM4XmcLDrqGZCyHs3wgH856BXyfyr4/" + id;
            fs.writeFileSync(SYSTEM.path + "systems/links/" + id, JSON.stringify({ mode: "resetpwd", id, sysId, userId: row.userId, email: data.email, oId: data.oId, datetime: new Date().toISOString() }));
            module_core_email({
                from: "Aimagin Connect <connect@aimagin.com>",
                to: data.email,
                subject: "Password reset link",
                html: "<p>Hello " + data.email + ", </p><div>You have reset your password " + (SYSTEM.serverType == "on-site" ? "" : ('(Organization ID: ' + data.oId + ')')) + ". Please click this <a href='" + link + "'>confirmation link</a> to reset your password.<br> If you don't want to reset your password, please ignore this email. This confirmation link is valid for 15 minutes.</div>"
            });
        }
    });
    db.close();
}

function module_main_resetPassword(sysId, data, callback) {
    let db = new sqlite3.Database(SYSTEM.path + "systems/" + sysId + "/db/system.db");
    let pwd = module_core_random(16);
    db.run("UPDATE tbl000 SET password = ? WHERE userId  = ?", [module_core_hash(pwd), data.userId], (err) => {
        if (err == null) {
            fs.unlinkSync(SYSTEM.path + "systems/links/" + data.id);
            module_core_email({
                from: "Aimagin Connect <connect@aimagin.com>",
                to: data.email,
                subject: "Your new password",
                html: "<p>Hello " + data.email + ", </p><div>You have reset your password " + (SYSTEM.serverType == "on-site" ? "" : ('(Organization ID: ' + data.oId + ')')) + ".<br>Your new password is <b>" + pwd + "</b>.<br>URL:" + SYSTEM.origin + (SYSTEM.serverType == "on-site" ? "" : ('/' + data.oId)) + "</div>"
            });
        }
        callback(err);
    });
    db.close();
}

function module_main_genRegisterLink(sysId, data) {
    if (fs.existsSync(SYSTEM.path + "systems/" + sysId + "/db/system.db")) {
        let db = new sqlite3.Database(SYSTEM.path + "systems/" + sysId + "/db/system.db");
        db.get("SELECT userId FROM tbl000 WHERE lower(email) = ?", [data.email.toLowerCase()], function (err, row) {
            if (row == undefined) {
                let id = module_core_genId(128);
                let link = SYSTEM.origin + "/yzbtpnaXb2NKC8YTq8naEqL4TcbLBNCQ94VM4XmcLDrqGZCyHs3wgH856BXyfyr4/" + id;
                fs.writeFileSync(SYSTEM.path + "systems/links/" + id, JSON.stringify({ mode: "register", id, sysId, server: SYSTEM.serverType, email: data.email, oId: data.oId, datetime: new Date().toISOString() }));
                module_core_email({
                    from: "Aimagin Connect <connect@aimagin.com>",
                    to: data.email,
                    subject: "Account registration link",
                    html: "<p>Hello " + data.email + ", </p><div>You have make a request for an account registration " + (SYSTEM.serverType == "on-site" ? "" : ('(Organization ID: ' + data.oId + ')')) + ". Please click this <a href='" + link + "'>confirmation link</a> to register.<br> If you don't want to register, please ignore this email. This confirmation link is valid for 15 minutes.</div>"
                });
            } else {
                module_core_email({
                    from: "Aimagin Connect <connect@aimagin.com>",
                    to: data.email,
                    subject: "Account already registered",
                    html: "<p>Hello " + data.email + ", </p><div>The user with email <b>" + data.email + "</b> has already been registered " + (SYSTEM.serverType == "on-site" ? "" : ('(Organization ID: ' + data.oId + ')')) + ". Please log in with your account.</div>"
                });
            }
        });
        db.close();
    }
}

function module_main_register(sysId, data) {
    fs.unlinkSync(SYSTEM.path + "systems/links/" + data.id);
    if (fs.existsSync(SYSTEM.path + "systems/" + sysId + "/db/system.db")) {
        let db = new sqlite3.Database(SYSTEM.path + "systems/" + sysId + "/db/system.db");
        db.get("SELECT userId FROM tbl000 WHERE lower(email) = ?", [data.email.toLowerCase()], function (err, row) {
            if (row == undefined) {
                db.run("INSERT INTO tbl000 (userId, email, password, firstname, lastname, department, mobile, line, page, roleId, status, approval, time_created) VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?)",
                    ["userId" + module_core_genId(16), data.email, module_core_hash(data.password), data.firstname, data.lastname, data.department, data.mobile, data.line, "0", "", "0", "0", new Date().toISOString()], (err) => {
                        if (err == null) {
                            module_core_email({
                                from: "Aimagin Connect <connect@aimagin.com>",
                                to: data.email,
                                subject: "Account registered",
                                html: "<p>Hello " + data.email + ", </p><div>You account has been registered " + (SYSTEM.serverType == "on-site" ? "" : ('(Organization ID: ' + data.oId + ')')) + ".<br>You can log in with the email <b>" + data.email + "</b> after your request is approved by administrators.</div>"
                                    + "<p><b>Account information</b></p><p><b>Email</b>: " + data.email + "</p><p><b>Password</b>: " + data.password + "</p><p><b>Firstname</b>: " + data.firstname + "</p><p><b>Lastname</b>: " + data.lastname + "</p><p><b>Department</b>: " + data.department + "</p><p><b>Mobile</b>: " + data.mobile + "</p><p><b>Line UID</b>: " + data.line + "</p>"
                            });
                        }
                    });
            } else {
                module_core_email({
                    from: "Aimagin Connect <connect@aimagin.com>",
                    to: data.email,
                    subject: "Account already registered",
                    html: "<p>Hello " + data.email + ", </p><div>The user with email <b>" + data.email + "</b> has already been registered " + (SYSTEM.serverType == "on-site" ? "" : ('(Organization ID: ' + data.oId + ')')) + ".</div>"
                });
            }
            db.close();
        });
    }
}

function module_main_post(system) {
    if (system.qpath[0] == 'esp32-uploadBinary' || system.qpath[0] == 'esp32-getssi') {
        let body = '';
        system.req.on('data', function (chunk) {
            body += chunk;
        }).on('end', function (e) {
            module_auth_authenticate(system, (status, err, cookie) => {
                if (system.qpath[0] == 'esp32-getssi') {
                    module_esp32('getSSIDs')(system, {}, (data) => {
                        module_main_responseJson(system.res, true, data, cookie);
                    });
                } else {
                }
            });
        })
        return;
    }


    module_core_parsePost(system, function (err, fields, files) {
        if (err != null) {
            module_main_responseJson(system.res, false, {
                ref: "err_3"
            });
            return;
        }
        if (files != null) {
            for (let n in files) {
                if (files[n].path == undefined) {
                    files[n].path = files[n].filepath;
                }
                if (files[n].hash == null) {
                    files[n].hash = files[n].newFilename;
                }
            }
        }
        if (fields._ == "login") {
            module_auth_logIn(system, fields.oId, fields.username, fields.password, (status, err, cookie) => {
                if (status) {
                    cookie.push('AIMAGIN_ACCESS_TYPE=private;');
                    module_main_responseJson(system.res, true, {}, cookie);
                } else {
                    module_main_responseJson(system.res, false, {
                        ref: err
                    }, cookie);
                }
            });
        } else if (fields._ == "init") {
            module_auth_authenticate(system, (status, err, cookie) => {
                if (system.user.userId == "Public") {
                    module_main_responseJson(system.res, true, {
                        user: null,
                        role: null,
                        webApp: null,
                        webAppMenu: null,
                        conf: JSON.parse(fs.readFileSync(system.sysPath + "/conf/conf.json"))
                    });
                } else {
                    if (status == true) {
                        module_main_getInitInfo(system);
                    } else {
                        module_main_errJson(err);
                    }
                }
            });
        } else if (fields["$_amgCnt_formData_submit_token_$"] != undefined) {
            module_auth_authenticate(system, (status, err, cookie) => {
                /*check permission*/
                module_sqlite("formDataReceived")(system, {
                    err,
                    fields,
                    files
                }, (data) => {
                    module_main_responseJson(system.res, true, data);
                });
            });
        } else if (fields["<amgCnt_upload_token>"] != undefined) {
            if (fields["<amgCnt_upload_file>"] != undefined) {
                module_auth_authenticate(system, (status, err, cookie) => {
                    module_server("6")(system, {
                        err,
                        fields,
                        files
                    }, (data) => {
                        module_main_responseJson(system.res, true, data);
                    });
                });
            } else if (fields["<amgCnt_upload_zip>"] != undefined) {
                module_main_responseJson(system.res, true, { path: files.file.path, hash: files.file.hash });
            } else if (fields["<amgCnt_upload_logo>"] != undefined) {
                module_auth_authenticate(system, (status, err, cookie) => {
                    module_setting(fields.mode)(system, {
                        err,
                        fields,
                        files
                    }, (data) => {
                        module_main_responseJson(system.res, true, data);
                    });
                });
            } else {
                module_auth_authenticate(system, (status, err, cookie) => {
                    module_drive("upload")(system, {
                        err,
                        fields,
                        files
                    }, (data) => {
                        module_main_responseJson(system.res, true, data);
                    });
                });
            }

        } else {
            module_auth_authenticate(system, function (status, err, cookie) {
                if (status) {
                    let moduleFn = null;
                    switch (fields._[0]) {
                        case "sqlite":
                            moduleFn = module_sqlite(fields._[1]);
                            break;
                        case "server":
                            moduleFn = module_server(fields._[1]);
                            break; /*server only*/
                        case "esp32":
                            moduleFn = module_esp32(fields._[1]);
                            break; /*esp32 only*/
                        case "clickhouse":
                            moduleFn = module_clickhouse(system, fields._[1]);
                            break;
                        case "mariadb":
                            moduleFn = module_mariadb(system, fields._[1]);
                            break;
                        case "drive":
                            moduleFn = module_drive(fields._[1]);
                            break;
                        case "python":
                            moduleFn = module_python(system, fields._[1]);
                            break;
                        case "datacleaning":
                            moduleFn = module_datacleaning(fields._[1]);
                            break;
                        case "file":
                            moduleFn = module_file(fields._[1]);
                            break;
                        case "setting":
                            moduleFn = module_setting(fields._[1]);
                            break;
                        case "webservice":
                            moduleFn = module_webservice(fields._[1]);
                            break;
                        case "view":
                            moduleFn = module_view(fields._[1]);
                            break;
                        case "exec":
                            moduleFn = module_exec(fields._[1]);
                            break;
                        case "iot":
                            moduleFn = module_iot(fields._[1]);
                            break;                            
                        case "conf":
                            switch (fields._[1]) {
                                case "backupset": moduleFn = function (system, data, callback) {
                                    fs.writeFileSync(system.sysPath + "/conf/backup.json", JSON.stringify(data), "utf8");
                                    callback({ data: JSON.parse(fs.readFileSync(system.sysPath + "/conf/backup.json", "utf8")) });
                                }
                                    break;
                                case "backupget": moduleFn = function (system, data, callback) {
                                    if (!fs.existsSync(system.sysPath + "/conf/backup.json")) {
                                        callback({ data: {} });
                                    } else {
                                        callback({ data: JSON.parse(fs.readFileSync(system.sysPath + "/conf/backup.json", "utf8")) });
                                    }
                                }
                                    break;
                            }
                            break;
                        case "backup":
                            moduleFn = module_backup(fields._[1]);
                            break;
                    };
                    if (moduleFn != null) {
                        moduleFn(system, fields._data, function (data, statusCode) {
                            module_main_responseJson(system.res, true, data, cookie, statusCode);
                        });
                    } else {
                        module_main_responseJson(system.res, false, {
                            ref: "err_3"
                        }, cookie);
                    }
                } else {
                    module_main_responseJson(system.res, false, {
                        ref: err
                    }, cookie);
                }
            });
        }
    });
}

(function () {
    if (fs.existsSync('./images_test')) {
        let imgs = fs.readdirSync('./images_test');
        let i = 0;
        module_main_esp32 = function (req, res, qpath) {
            if (req.method == "GET") {
                if (qpath[0] == 'capture') {
                    res.writeHead(200, {
                        "Content-Type": 'image/jpg'
                    });
                    fs.createReadStream('./images_test/' + imgs[i++ % imgs.length]).pipe(res);
                } else if (qpath[0] == 'stream') {
                    res.writeHead(200, {
                        "Content-Type": 'multipart/x-mixed-replace; boundary=myBoundary'
                    });
                    let n = setInterval(() => {
                        let buffer = fs.readFileSync('./images_test/' + imgs[i++ % imgs.length]);
                        res.write('--myBoundary\r\n')
                        res.write('Content-Type:image/jpg\r\nContent-Length:' + buffer.byteLength + '\r\n\r\n');
                        res.write(buffer);
                        res.write("\r\n")
                    }, 20);
                    req.on('close', function (e) {
                        console.log('close');
                        clearInterval(n);
                    });
                } else if (qpath[0] == "status") {
                    res.writeHead(200, {
                        "Content-Type": 'application/json'
                    });
                    res.end(JSON.stringify({
                        framesize: 0
                    }));
                } else if (qpath[0] == "save") {
                    console.log('save')
                    res.writeHead(200, {
                        "Content-Type": 'application/json'
                    });
                    res.end(JSON.stringify({
                        framesize: 0
                    }));
                }
            }
        }
    }
})();

module.exports.upgrade = function (req, res, head) {
    //for websocket.      
};

let zoomlevel = {}
for (let i = 0; i < 23; i++) {
    zoomlevel[i] = Math.pow(2, i);
}

const RESTFUL = {
    etl_time: "(SELECT max(ETL_time) FROM Thisis_MD_SHIPALL where ABORT_FLAG_DESC = 'มีทะเบียน' and USE_TYPE_NAME like '%ประมง%' and toFloat64OrZero(TGROSS) >= 30)",
};


(function init_restful() {
    //about data
    if(!fs.existsSync("./data"))    return;
    (function update_data() {
        try {
            let last_updated = [];
            fs.readdirSync("./data").forEach(function (d) {                
                if(d=="last_updated.txt")   return;
                last_updated.push(d);
                let [api, v] = d.replace(".json", "").split("-");
                try {
                    if (RESTFUL[api]) {
                        if (RESTFUL[api].v < v) {                            
                            RESTFUL[api] = { v:v, rows:JSON.parse(fs.readFileSync("./data/" + d, "utf8"))};
                            last_updated.push(api+" - "+RESTFUL[api].v+" - New Updated Values. " + RESTFUL[api].rows.length) ;
                        } else {
                            last_updated.push(api + " - " + RESTFUL[api].v +" - No Updated Values. " + RESTFUL[api].rows.length);
                        }
                    } else {
                        RESTFUL[api] = { v:v, rows:JSON.parse(fs.readFileSync("./data/" + d, "utf8"))};
                        last_updated.push(api+" - "+RESTFUL[api].v+" - New Updated Values. " + RESTFUL[api].rows.length);
                    }
                } catch(e) {
                    last_updated.push(api+" - Failed to Updated Values.");
                }
            });
            fs.writeFileSync("./data/last_updated.txt", last_updated.join("\r\n"), "utf8");
        } catch (e) {
            console.log(e);
        }
        setTimeout(update_data, 10 * 1000);
    })();
})();

let restful = 0
module.exports.connect = function (req, res) {
    let q = url.parse(req.url, true);
    let qpath = q.pathname.split("/");
    res.setHeader("Access-Control-Allow-Origin", SYSTEM.origin);
    qpath.shift(); /*remove first empty string*/
    if (req.method == "GET" && (qpath[0] == "proxy_tunnel")) {
        let domain = qpath[1];
        let port = qpath[2];
        let path = req.url

        let pathA = path.split('/');
        if (pathA.length) { pathA.splice(0, 1); }
        if (pathA.length) { pathA.splice(0, 1); }
        if (pathA.length) { pathA.splice(0, 1); }
        if (pathA.length) { pathA.splice(0, 1); }
        path = `/${pathA.join('/')}`;

        /*var options = {
          host: "fgds.dmcr.go.th",
          port: 8000,
          path: '/geoserver/dmcr_protection/wms?service=WMS&request=GetMap&layers=dmcr_protection%3Admcr_underwaterpinnacle&styles=&format=image/png&transparent=false&version=1.1.1&width=256&height=256&srs=EPSG%3A3857&bbox=10018754.171394622,0,11271098.44281895,1252344.2714243277',
        };*/
        var options = {
            host: domain,
            port: port,
            path: path,
        };
        var connector = http.request(options, function (connector_res) {
            connector_res.pipe(res, { end: true });//tell 'response' end=true
        });
        req.pipe(connector, { end: true });
    } else if (req.method == "POST" && (qpath[0] == "restful" || qpath[0] == "restful2")) {
        res.setHeader("Access-Control-Allow-Origin", '*');
        let resp = {
            err: false,
            errMsg: '',
            query: {},
            rows: [],
            ver: `1.0.7`
        };
        let string_err_hides = [new RegExp(SYSTEM.sysId + '.', "g")];
        let target_accepts = [
            'Nmea_raw', 'VMS_info', 'VMS_info_all', 
            'VMS_complete_info_all_ships', 'VMS', 'tracking', 'ship_info', 
            'Thailand_area', 'Nmea_raw_last', 'Nmea_THISIS_raw', 'Nmea_THISIS_raw_last', 
            'IHS_info_last', 'VMS_complete_info_all_ships2', 'Nmea_raw_last2', 'Nmea_THISIS_raw_last2'
        ];

        if (qpath[0] != "restful2") {
            if (restful > 20) {
                resp.err = true;
                resp.errMsg = 'exceed query limit.'
                res.writeHead(200, { 'Content-Type': 'application/json' });
                res.end(JSON.stringify(resp));
                return;
            }
            restful += 1;
        }

        console.log(restful);

        fn_error_resp = (msg) => {
            resp.err = true;
            resp.errMsg = msg;
            resp.rows = [];
            res.end(JSON.stringify(resp));
            if (qpath[0] != "restful2") { restful -= 1; }
        }
        fn_query_err = (name, lv) => {
            if (lv == 1) {
                if (typeof resp.query[name] == 'undefined' || resp.query[name] == '' || resp.query[name] == null) {
                    return true;
                } else {
                    resp.query[name] = resp.query[name].replace(/;/g, ' BLOCK ');
                    return false;
                }
            } else if (lv == 2) {
                if (typeof resp.query[name] == 'undefined') {
                    return true;
                } else {
                    resp.query[name] = resp.query[name].replace(/;/g, ' BLOCK ');
                    return false;
                }
            } return true;
        }
        fn_db_err = (errMsg) => {
            resp.err = true;
            resp.errMsg = errMsg;
            string_err_hides.map(re => {
                resp.errMsg = resp.errMsg.replace(re, '');
            })
        }

        req.on('data', function (chunk) {
            resp.query += chunk;
        }).on('end', function (e) {
            try {
                let n = resp.query.indexOf('{');
                let n2 = resp.query.lastIndexOf('}') + 1;
                resp.query = resp.query.slice(n, n2);
                resp.query = JSON.parse(resp.query);

                const now = new Date();
                const m = now.getMonth() + 1;
                const d = now.getDate();
                const h = now.getHours();
                fs.appendFileSync(`./process/API_${m}_${d}_${h}.log`, `${resp.query.target}|${qpath[0]}|${new Date().toISOString()}\r\n`);

                if (fn_query_err('key', 1) || resp.query.key != 'gZ37DyAdObZ7TT28uNGzEh4Jd2uArLf4') {
                    fn_error_resp("Wrong API key");
                } else if (fn_query_err('target', 1)) {
                    fn_error_resp("Wrong query target");
                } else if (!target_accepts.includes(resp.query.target)) {
                    fn_error_resp("Wrong query target");
                } else {
                    try {
                        const clickhouse = new ClickHouse({
                            url: 'http://localhost',
                            port: 8123,
                            password: SYSTEM.clickhousepwd,
                            format: "json", // "json" || "csv" || "tsv"
                            database: SYSTEM.sysId,
                        });
                        let limit = 5000;

                        if (resp.query.target == 'Nmea_THISIS_raw_last2') {
                            let d_now = new Date()
                            let d_start = new Date()
                            d_now.setTime(d_now.getTime() + (7 * 60 * 60 * 1000))
                            let age_minute = 7 * 60 * 60 * 1000
                            if (resp.query.age) {
                                age_minute -= +resp.query.age ? +resp.query.age : 0
                            } else {
                                age_minute -= 1 * 60 * 1000
                            }
                            d_start.setTime(d_start.getTime() + (age_minute))
                            d_start = d_start.toISOString()
                            d_start0 = d_start
                            d_start = d_start.replace('T', ' ')
                            d_start = d_start.slice(0, d_start.indexOf('.'))
                            d_now = d_now.toISOString()
                            d_now = d_now.replace('T', ' ')
                            d_now = d_now.slice(0, d_now.indexOf('.'))
                            resp.query.between = { from: d_start, to: d_now }
                            clickhouse.query(`SELECT *
                            FROM THISIS_NMEA_VIEW 
                            WHERE block_c >= '${d_start0}' 
                            ORDER BY block_c DESC 
                            LIMIT ${limit}`).exec(function (err, rows) {
                                if (err != null) {
                                    fn_db_err(err.toString());
                                } else {
                                    resp.rows = rows;
                                }
                                res.writeHead(200, { 'Content-Type': 'application/json' });
                                res.end(JSON.stringify(resp));
                                if (qpath[0] != "restful2") { restful -= 1; }
                            });
                        } else if (resp.query.target == 'Nmea_THISIS_raw_last') {
                            resp.query.between = { from: null, to: null }                            
                            try {
                                resp.rows = RESTFUL['Nmea_THISIS_raw_last'].rows;
                            } catch (e) {
                                resp.rows = []
                            }
                            res.writeHead(200, { 'Content-Type': 'application/json' });
                            res.end(JSON.stringify(resp));
                            if (qpath[0] != "restful2") { restful -= 1; }
                        } else if (resp.query.target == 'Nmea_THISIS_raw') {
                            if (fn_query_err('from', 1)) {
                                fn_error_resp("Wrong query start date");
                            } else if (fn_query_err('to', 1)) {
                                fn_error_resp("Wrong query end date");
                            } else {
                                clickhouse.query(`SELECT *
                                    FROM THISIS_NMEA
                                    WHERE replaceRegexpAll(block_c, '(t)|(T)', ' ') >= '${resp.query.from}'
                                    AND replaceRegexpAll(block_c, '(t)|(T)', ' ') <= '${resp.query.to}'
                                    ORDER BY block_c DESC
                                    LIMIT ${limit}`).exec(function (err, rows) {
                                    if (err != null) {
                                        fn_db_err(err.toString());
                                    } else {
                                        resp.rows = rows;
                                    }
                                    res.writeHead(200, { 'Content-Type': 'application/json' });
                                    res.end(JSON.stringify(resp));
                                    if (qpath[0] != "restful2") { restful -= 1; }
                                });
                            }
                        } else if (resp.query.target == 'Nmea_raw_last') {
                            let d_now = new Date()
                            let d_start = new Date()
                            d_now.setTime(d_now.getTime() + (7 * 60 * 60 * 1000))
                            let age_minute = 7 * 60 * 60 * 1000
                            if (resp.query.age) {
                                age_minute -= +resp.query.age ? +resp.query.age : 0
                            } else {
                                age_minute -= 1 * 60 * 1000
                            }
                            d_start.setTime(d_start.getTime() + (age_minute))
                            d_start = d_start.toISOString()
                            d_start0 = d_start
                            d_start = d_start.replace('T', ' ')
                            d_start = d_start.slice(0, d_start.indexOf('.'))
                            d_now = d_now.toISOString()
                            d_now = d_now.replace('T', ' ')
                            d_now = d_now.slice(0, d_now.indexOf('.'))
                            resp.query.between = { from: d_start, to: d_now }
                            clickhouse.query(`SELECT *
                            FROM HYDRO_NMEA_VIEW 
                            WHERE block_c >= '${d_start0}' 
                            ORDER BY block_c DESC 
                            LIMIT ${limit}`).exec(function (err, rows) {
                                if (err != null) {
                                    fn_db_err(err.toString());
                                } else {
                                    resp.rows = rows;
                                }
                                res.writeHead(200, { 'Content-Type': 'application/json' });
                                res.end(JSON.stringify(resp));
                                if (qpath[0] != "restful2") { restful -= 1; }
                            });
                        } else if (resp.query.target == 'Nmea_raw_last2') {
                            try {
                                resp.rows = RESTFUL['Nmea_raw_last'].rows;
                            } catch (e) {
                                resp.rows = []
                            }
                            res.writeHead(200, { 'Content-Type': 'application/json' });
                            res.end(JSON.stringify(resp));
                            if (qpath[0] != "restful2") { restful -= 1; }
                        } else if (resp.query.target == 'Nmea_raw') {
                            if (fn_query_err('from', 1)) {
                                fn_error_resp("Wrong query start date");
                            } else if (fn_query_err('to', 1)) {
                                fn_error_resp("Wrong query end date");
                            } else {
                                clickhouse.query(`SELECT *
                                    FROM HYDRO_NMEA
                                    WHERE replaceRegexpAll(block_c, '(t)|(T)', ' ') >= '${resp.query.from}'
                                    AND replaceRegexpAll(block_c, '(t)|(T)', ' ') <= '${resp.query.to}'
                                    ORDER BY block_c DESC
                                    LIMIT ${limit}`).exec(function (err, rows) {
                                    if (err != null) {
                                        fn_db_err(err.toString());
                                    } else {
                                        resp.rows = rows;
                                    }
                                    res.writeHead(200, { 'Content-Type': 'application/json' });
                                    res.end(JSON.stringify(resp));
                                    if (qpath[0] != "restful2") { restful -= 1; }
                                });
                            }
                        } else if (resp.query.target == 'VMS_info') {
                            if (fn_query_err('vessel_id', 1)) {
                                fn_error_resp("Wrong query VESSEL ID number");
                            } else {
                                clickhouse.query(`SELECT *
                                    FROM VMS_FISHERY_SHIP_MD
                                    WHERE Vessel_Id = '${resp.query.vessel_id}'
                                    ORDER BY Vessel_Id ASC
                                    LIMIT ${limit}`).exec(function (err, rows) {
                                    if (err != null) {
                                        fn_db_err(err.toString());
                                    } else {
                                        resp.rows = rows;
                                    }
                                    res.writeHead(200, { 'Content-Type': 'application/json' });
                                    res.end(JSON.stringify(resp));
                                    if (qpath[0] != "restful2") { restful -= 1; }
                                });
                            }
                        } else if (resp.query.target == 'VMS_info_all') {
                            clickhouse.query(`SELECT Vessel_Id,Ship_name
                                FROM VMS_FISHERY_SHIP_MD
                                ORDER BY Vessel_Id ASC`).exec(function (err, rows) {
                                if (err != null) {
                                    fn_db_err(err.toString());
                                } else {
                                    resp.rows = rows;
                                }
                                res.writeHead(200, { 'Content-Type': 'application/json' });
                                res.end(JSON.stringify(resp));
                                if (qpath[0] != "restful2") { restful -= 1; }
                            });
                        } else if (resp.query.target == 'VMS_complete_info_all_ships2') {
                            clickhouse.query(`
                                SELECT 
                                    A.SHIP_CODE Vessel_Id, A.THA_NAME Ship_name, B.Speed AS Speed, B.Heading AS Heading, B.LocatedAt AS LocatedAt, 
                                    B.Latitude AS Latitude, B.Longitude AS Longitude 
                                FROM (
                                    SELECT SHIP_CODE, THA_NAME FROM Thisis_MD_SHIPALL 
                                    WHERE ETL_time= ${RESTFUL.etl_time} AND ABORT_FLAG_DESC = 'มีทะเบียน' and USE_TYPE_NAME like '%ประมง%' and toFloat64OrZero(TGROSS) >= 30
                                ) A
                                INNER JOIN (
                                    SELECT * FROM VMS_FISHERY_SHIP_LOCATION 
                                    WHERE LocatedAt>= substr(toString(subtractDays(now('Asia/Bangkok'), 1)), 1, 10) AND VesselId || LocatedAt IN (
                                        SELECT VesselId || max(LocatedAt) FROM VMS_FISHERY_SHIP_LOCATION 
                                        WHERE LocatedAt>= substr(toString(subtractDays(now('Asia/Bangkok'), 1)), 1, 10) 
                                        GROUP BY VesselId
                                    )
                                ) B ON A.SHIP_CODE = B.VesselId
                            `).exec(function (err, rows) {
                                if (err != null) {
                                    fn_db_err(err.toString());
                                } else {
                                    resp.rows = rows;
                                }
                                res.writeHead(200, { 'Content-Type': 'application/json' });
                                res.end(JSON.stringify(resp));
                                if (qpath[0] != "restful2") { restful -= 1; }
                            });
                        } else if (resp.query.target == 'VMS_complete_info_all_ships') {
                            try {
                                if(resp.query.use_filter==1) {
                                    resp.rows = RESTFUL['VMS_complete_info_all_ships_filter'].rows;
                                } else {
                                    resp.rows = RESTFUL['VMS_complete_info_all_ships'].rows;
                                }    
                            } catch (e) {
                                resp.rows = []
                            }
                            res.writeHead(200, { 'Content-Type': 'application/json' });
                            res.end(JSON.stringify(resp));
                            if (qpath[0] != "restful2") { restful -= 1; }
                        } else if (resp.query.target == 'IHS_info_last') {
                            try {
                                resp.rows = RESTFUL['IHS_info_last'].rows;
                            } catch (e) {
                                resp.rows = []
                            }
                            res.writeHead(200, { 'Content-Type': 'application/json' });
                            res.end(JSON.stringify(resp));
                            if (qpath[0] != "restful2") { restful -= 1; }
                        } else if (resp.query.target == 'VMS') {
                            if (fn_query_err('vessel_id', 1)) {
                                fn_error_resp("Wrong query VESSEL ID number");
                            } else if (fn_query_err('from', 1)) {
                                fn_error_resp("Wrong query start date");
                            } else if (fn_query_err('to', 1)) {
                                fn_error_resp("Wrong query end date");
                            } else {
                                clickhouse.query(`SELECT *
                                    FROM VMS_FISHERY_SHIP_LOCATION
                                    WHERE VesselId = '${resp.query.vessel_id}'
                                    AND replaceRegexpAll(LocatedAt, '(t)|(T)', ' ') >= '${resp.query.from}'
                                    AND replaceRegexpAll(LocatedAt, '(t)|(T)', ' ') <= '${resp.query.to}'
                                    ORDER BY LocatedAt DESC
                                    LIMIT ${limit}`).exec(function (err, rows) {
                                    if (err != null) {
                                        fn_db_err(err.toString());
                                    } else {
                                        resp.rows = rows;
                                    }
                                    res.writeHead(200, { 'Content-Type': 'application/json' });
                                    res.end(JSON.stringify(resp));
                                    if (qpath[0] != "restful2") { restful -= 1; }
                                });
                            }
                        } else if (resp.query.target == 'Thailand_area') {
                            if (fn_query_err('imo_or_mmsi', 1)) {
                                fn_error_resp("Wrong query IMO/MMSI number");
                            } else if (fn_query_err('from', 1)) {
                                fn_error_resp("Wrong query start date");
                            } else if (fn_query_err('to', 1)) {
                                fn_error_resp("Wrong query end date");
                            } else {
                                clickhouse.query(`SELECT *
                                    FROM IHS_AIS_INFO_AREA
                                    WHERE (IMO = '${resp.query.imo_or_mmsi}' OR MMSI = '${resp.query.imo_or_mmsi}')
                                    AND replaceRegexpAll(MessageTimestamp, '(t)|(T)', ' ') >= '${resp.query.from}'
                                    AND replaceRegexpAll(MessageTimestamp, '(t)|(T)', ' ') <= '${resp.query.to}'
                                    ORDER BY IMO ASC, MMSI ASC, MessageTimestamp DESC
                                    LIMIT ${limit}`).exec(function (err, rows) {
                                    if (err != null) {
                                        fn_db_err(err.toString());
                                    } else {
                                        resp.rows = rows;
                                    }
                                    res.writeHead(200, { 'Content-Type': 'application/json' });
                                    res.end(JSON.stringify(resp));
                                    if (qpath[0] != "restful2") { restful -= 1; }
                                });
                            }
                        } else if (resp.query.target == 'tracking') {
                            if (fn_query_err('imo_or_mmsi', 1)) {
                                fn_error_resp("Wrong query IMO number");
                            } else if (fn_query_err('from', 1)) {
                                fn_error_resp("Wrong query start date");
                            } else if (fn_query_err('to', 1)) {
                                fn_error_resp("Wrong query end date");
                            } else {
                                clickhouse.query(`SELECT * 
                                        FROM IHS_AIS_INFO 
                                    WHERE (IMO = '${resp.query.imo_or_mmsi}' OR MMSI = '${resp.query.imo_or_mmsi}')
                                    AND replaceRegexpAll(MessageTimestamp, '(t)|(T)', ' ') >= '${resp.query.from}'
                                    AND replaceRegexpAll(MessageTimestamp, '(t)|(T)', ' ') <= '${resp.query.to}'
                                    ORDER BY IMO ASC, MMSI ASC, MessageTimestamp DESC
                                    LIMIT ${limit}`).exec(function (err, rows) {
                                    if (err != null) {
                                        fn_db_err(err.toString());
                                    } else {
                                        resp.rows = rows;
                                    }
                                    res.writeHead(200, { 'Content-Type': 'application/json' });
                                    res.end(JSON.stringify(resp));
                                    if (qpath[0] != "restful2") { restful -= 1; }
                                });
                            }
                        } else if (resp.query.target == 'ship_info') {
                            if (fn_query_err('imo_or_ihslr', 1)) {
                                fn_error_resp("Wrong query IMO/IHSLR number");
                            } else {
                                clickhouse.query(`SELECT *
                                    FROM IHS_SHIP_INFO
                                    WHERE IHSLRorIMOShipNo = '${resp.query.imo_or_ihslr}' 
                                    ORDER BY IHSLRorIMOShipNo ASC
                                    LIMIT ${limit}`).exec(function (err, rows) {
                                    if (err != null) {
                                        fn_db_err(err.toString());
                                    } else {
                                        resp.rows = rows;
                                    }
                                    res.writeHead(200, { 'Content-Type': 'application/json' });
                                    res.end(JSON.stringify(resp));
                                    if (qpath[0] != "restful2") { restful -= 1; }
                                });
                            }
                        } else if (resp.query.target == 'custom') {
                            if (fn_query_err('key2', 1) || resp.query.key2 != 'krh3S89W7sGDzdmFSMhJZFaU2PkQZ5UV63NNMMAJ') {
                                fn_error_resp("Wrong key2");
                            } else {
                                if (resp.query.key3 == 'A') {
                                    //let db = new sqlite3.Database(path.path, sqlite3.OPEN_READONLY);

                                } else if (resp.query.key3 == 'B') {
                                    clickhouse.query(`SELECT * FROM ${resp.query.sql.replace(/;/g, '')}`).exec(function (err, rows) {
                                        if (err != null) {
                                            fn_db_err(err.toString());
                                        } else {
                                            resp.rows = rows;
                                        }
                                        res.writeHead(200, { 'Content-Type': 'application/json' });
                                        res.end(JSON.stringify(resp));
                                        if (qpath[0] != "restful2") { restful -= 1; }
                                    });
                                } else {
                                    fn_error_resp("Wrong key3");
                                }
                            }
                        } else { fn_error_resp("Not found query target"); }
                    } catch (e) { fn_error_resp(e.message); }
                }
            } catch (e) {
                fn_error_resp("Wrong query data");
            }
        });
    } else if (req.method == "GET" && (qpath[0] == "include" || qpath[0] == "favicon.ico" || qpath[0] == "text")) {
        if (qpath[0] == "text") {
            res.writeHead(200, {
                "Content-Type": 'text/plain'
            });
            if (fs.existsSync(SYSTEM.path + 'core/text/' + qpath[1])) {
                fs.createReadStream(SYSTEM.path + 'core/text/' + qpath[1]).pipe(res);
            } else {
                res.end('');
            }
        } else {
            module_main_include(qpath, res);
        }
    } else if ((req.method == "GET" || req.method == "POST") && qpath[0] == "api") {
        module_webservice("request")(req, res, (status) => {
            if (status == false) {
                module_main_responseGzip(res, "./core/client/err_404.html.gz", { "Content-Type": "text/html" });
            }
        });
    } else if (req.method == "POST" && qpath[0] == 'bN3tAu9ywPqzLemL3JMeEnMpbNaqnHULWB2gVv6P33axrse9v4hymNYAuzMnqUvP') {
        module_mobile(req, res);
    } else if (qpath[0] == 'wZurnPy3sBQg54vvna3SK6ayVNjSFstWYZfkbXzDkXFh29bjYfRt3mvU8EPpKbSe') {
        let qdata = url.parse(req.url, true).query;
        let img_path = `./systems/${qdata['sysId']}/drive/data/${qdata['loc']}/${qdata['dt']}/${qdata['z']}_${((+qdata['y']) * zoomlevel[qdata['z']] + (+qdata['x']))}.png.gz`;
        if (fs.existsSync(img_path)) {
            res.writeHead(200, { 'Content-type': 'image/png', 'Content-Encoding': "gzip" })
            fs.createReadStream(img_path).pipe(res);
        } else {
            res.end(null);
        }
    } else if (qpath[0] == 'uwvzD6CypyNyDq5FzVrR4CFbDadtXHDycDUVG3gvmu6HeV8eJFMzzNbEFEWq2RZU') {
        qpath.shift();
        module_main_esp32(req, res, qpath);
    } else if (qpath[0] == 'U4tfcUMqnwMxWrGLwNZKdWDMx2XGYfGgb2LzpZJdY6dVdQQ9ReJ377aJEBht6Rkk') {
        res.setHeader("Access-Control-Allow-Origin", '*');
        let mode = qpath[1];
        let body = '';
        req.on('data', function (chunk) {
            body += chunk;
        }).on('end', function (e) {
            try {
                let data = JSON.parse(body);
                let sysId = module_auth_readSysId(data.id)
                if (sysId == null) {
                    res.end("Failure");
                    return;
                }
                if (mode == "insert") {
                    module_sqlite("API.insert")({
                        sysPath: SYSTEM.path + "systems/" + sysId,
                        sysId
                    }, data, (r) => {
                        if (r.err) {
                            res.end("Failure: " + r.err);
                        } else {
                            res.end("Success");
                        }
                    });
                } else {
                    res.end("Failure: Incorrect mode");
                }
            } catch (e) {
                res.end("Failure: Incorrect format");
            }
        })
    } else if (req.method == "POST" && qpath[0] == 'AAEZfmp86FaHLQVttTPTfKcgUsEkhc8w3pxHeX8FN8dFC4ST7Zx7ScpRng9LpYSy') {
        module_core_parsePost({ req, res }, function (err, fields, files) {
            let sysId = SYSTEM.serverType == "on-site" ? SYSTEM.sysId : fs.existsSync(SYSTEM.path + "systems/oId/" + fields.oId) ? fs.readFileSync(SYSTEM.path + "systems/oId/" + fields.oId, "utf8") : null;
            if (sysId != null) {
                switch (qpath[1]) {
                    case "resetpwd": module_main_genResetPasswordLink(sysId, fields); break;
                    case "register": module_main_genRegisterLink(sysId, fields); break;
                    default: ;
                }
            }
            res.end(null);
        });
    } else if (qpath[0] == 'yzbtpnaXb2NKC8YTq8naEqL4TcbLBNCQ94VM4XmcLDrqGZCyHs3wgH856BXyfyr4') {
        if (fs.existsSync(SYSTEM.path + "systems/links/" + qpath[1])) {
            if (req.method == "GET") {
                let data = JSON.parse(fs.readFileSync(SYSTEM.path + "systems/links/" + qpath[1]));
                if (new Date() - new Date(data.datetime) < 15 * 60 * 1000) {
                    if (data.mode == "resetpwd") {
                        module_main_resetPassword(data.sysId, data, function (err) {
                            if (err) {
                                module_main_responseGzip(res, "./core/client/err_link.html.gz", { "Content-Type": "text/html" });
                            } else {
                                module_main_responseGzip(res, "./core/client/success_resetpwd.html.gz", { "Content-Type": "text/html" });
                            }
                        });
                    } else if (data.mode == "register") {
                        module_main_responseGzip(res, "./core/client/register.html.gz", { "Content-Type": "text/html" });
                    } else {
                        module_main_responseGzip(res, "./core/client/err_404.html.gz", { "Content-Type": "text/html" });
                    }
                } else {
                    module_main_responseGzip(res, "./core/client/err_link.html.gz", { "Content-Type": "text/html" });
                }
            } else if (req.method == "POST") {
                module_core_parsePost({ req, res }, function (err, fields, files) {
                    if (fields.mode == "0") {
                        fs.createReadStream(SYSTEM.path + "systems/links/" + qpath[1]).pipe(res);
                    } else if (fields.mode == "1") {
                        module_main_register(fields.sysId, fields);
                        res.end(JSON.stringify({ status: true }));
                    } else {
                        res.end(null);
                    }
                });
            }
        } else {
            module_main_responseGzip(res, "./core/client/err_link.html.gz", { "Content-Type": "text/html" });
        }
    } else if(qpath[0]=='wZiTxuNLVLkmnfnzXNVyTB3K7XXb9SSioWQAfAlV7aVBTOAobCbsov1ULwCxInT5') {
        qpath.shift();
        module_iot("snooze_alert")(req, res, qpath);
    } else if(req.method=="POST" && qpath[0]=='8V4v5rmBehMXqKTv8eBjWKQQjaDJgSxH4xCmTjYzzzxE5L9knEWfgvYJzy9RfNrU') {
        qpath.shift();
        switch(qpath[0]) {
            case "296UpgFQAJtpvjWR25trnRzeZpnbp2yY4q2b2dvW79gq6bNNZZmnurpg2cdV4hBS": module_server("line")(req, res, qpath); break;
            case "p6BzkUc8hsPbumC7mWCye7jLptJMXDKrDc74JStz6tpyJCqHN5xEPynXyVTesSB5": module_server("email")(req, res, qpath);break;
            case "BgCh2WVNXFkT5aKkSLw6dc8eMLR4jvs4NnLcUmjTGqGJHX9xwgh4KAEjQGAsUgwZ": module_server("line_webhook")(req, res, qpath);
            default: return res.end(null);
        }
    } else if(req.method === "POST" && qpath[0] === "bv4ql312j7b2Y1bs1691378352719Y2ewQvfhMj5mggNp1691378352719hGfzfqvb6S1tlJVc"){
        // Backup System
        /// for send mail backup status.
        const func = module_backup("send_mail_status");
        if(func){
            module_core_parsePost({req,res},function (err, fields, files){
                func({req,res,sysId: SYSTEM.sysId},fields, function (data, statusCode) {
                    module_main_responseJson(res, true, data, undefined, statusCode);
                });
            });
        }else{
            res.end(null);
        }
    } else {
        let system = {
            req,
            res,
            pathname: q.pathname,
            qpath,
            qdata: q.query,
            sysPath: SYSTEM.sysPath,
            sysId: SYSTEM.sysId,
            oId: null,
            cookie: module_core_parseCookie(req.headers.cookie),
            cookie0: req.headers.cookie
        };
        if (req.method == "GET") {
            if (SYSTEM.serverType == "cloud") {
                /*server only*/
                if (system.qpath.length == 1) {
                    system.defaultPage = true;
                    if (system.qpath[0] != "") {
                        system.oId = system.qpath[0];
                        system.oIdInPath = true;
                    }
                } else if (system.qpath.length == 2) {
                    system.oId = system.qpath[0];
                    system.publicPath = system.qpath[1];
                    system.public = true;
                    system.qdata = {};
                } else {
                    module_main_responseGzip(res, "./core/client/err_404.html.gz", {
                        "Content-Type": "text/html",
                        //"Set-Cookie":system.cookie0
                    });
                    return;
                }
                /*server only*/
            } else {
                if (system.qpath.length == 1) {
                    if (system.qpath[0] == "") {
                        system.defaultPage = true;
                    } else {
                        system.publicPath = system.qpath[0];
                        system.public = true;
                    }
                } else {

                    module_main_responseGzip(res, "./core/client/err_404.html.gz", {
                        "Content-Type": "text/html",
                        //"Set-Cookie":system.cookie0
                    });
                    return;
                }
            }
            module_main_get(system);
        } else if (req.method == "POST") {
            module_main_post(system);
        } else {
            res.end(null);
        }
    }
};

console.log("SERVER START");