module_sqlite = function (mode) {
    let fn = null;
    switch (mode) {
        case "run":
            fn = module_sqlite_run;
            break;
        case "all":
            fn = module_sqlite_all;
            break;
        case "get":
            fn = module_sqlite_get;
            break;
        case "formDataReceived":
            fn = module_sqlite_formData_received;
            break;
        case "updateTable":
            fn = module_sqlite_updateTable;
            break;
        case "create":
            fn = module_sqlite_create;
            break;
        case "read":
            fn = module_sqlite_read;
            break;
        case "update":
            fn = module_sqlite_update;
            break;
        case "delete":
            fn = module_sqlite_delete;
            break;
        case "API.insert":
            fn = module_sqlite_insertAPI;
            break;
        case "updateDataDictionaryFormat":
            fn = module_sqlite_updateDataDictionaryFormat;
            break;
        case "updateDatabase":
            fn = module_sqlite_updateDatabase;
            break;
    }
    return fn;
}

function params(user, params, hashIndex) {
    let keywords = {
        "$_userId_$": user.userId,
        "$_datetime_iso_$": new Date().toISOString(),
        "$_datetime_ms_$": new Date().getTime()
    };
    let p = params.map(d => keywords[d] != undefined ? keywords[d] : d);
    if (hashIndex != undefined && typeof hashIndex.forEach == "function") {
        hashIndex.forEach(function (d) {
            p[d] = SYSTEM.module.core.hash(p[d]);
        });
    }
    return p;
}

module_sqlite_run = function (system, data, callback) {
    const begin_execute_time = new Date();
    let configLog={apiName:'module_sqlite_run', config: JSON.stringify(data)};

    module_sqlite_path(system, data, function (path) {
        if (path.err) {
            try{
                const end_execute_time = new Date();
                var diff = (end_execute_time-begin_execute_time);
                configLog.processTime=diff;
                configLog.errMsg = path.err;
                module_core_recordRequestLog(system, system.user.email, 'sqlite', configLog);
            }catch(e){}
            finally{
                callback({
                    err: path.err,
                    snapshot: path.snapshot,
                    errMsg: path.err
                });
            }
        } else {
            let db = new sqlite3.Database(path.path);
            db.run(data.sql, params(system.user, data.params, data.hashIndex), (err) => {
                try{
                    const end_execute_time = new Date();
                    var diff = (end_execute_time-begin_execute_time);
                    configLog.processTime=diff;
                    configLog.errMsg = path.err;
                    module_core_recordRequestLog(system, system.user.email, 'sqlite', configLog);
                }catch(e){}
                finally{
                    callback({
                        err: err,
                        snapshot: path.snapshot,
                        errMsg: err != null ? err.message : undefined,
                    });
                }
            });
            db.close();
        }
    });
}

let MAX_STRICT_ALL = 0;
module_sqlite_all = function (system, data, callback) {
    const begin_execute_time = new Date();
    let configLog={apiName:'module_sqlite_all', config: JSON.stringify(data)};
 
    module_sqlite_path(system, data, function (path) {
        if (path.err) {
            try{
                const end_execute_time = new Date();
                var diff = (end_execute_time-begin_execute_time);
                configLog.processTime=diff;
                configLog.errMsg = path.err;
                module_core_recordRequestLog(system, system.user.email, 'sqlite', configLog);
            }catch(e){}
            finally{
                callback({
                    err: path.err,
                    snapshot: path.snapshot,
                    errMsg: path.err,
                    rows: null
                });
            }
        } else {
            if (data.strict == true) {
                if (MAX_STRICT_ALL < 1) {
                    MAX_STRICT_ALL++;
                    let db = new sqlite3.Database(path.path, sqlite3.OPEN_READONLY);
                    db.all(module_core_selectOnlySql(data.sql), params(system.user, data.params), (err, rows) => {
                        MAX_STRICT_ALL--;
                        try{
                            const end_execute_time = new Date();
                            var diff = (end_execute_time-begin_execute_time);
                            configLog.processTime=diff;
                            configLog.errMsg = path.err;
                            module_core_recordRequestLog(system, system.user.email, 'sqlite', configLog);
                        }catch(e){}
                        finally{
                            callback({
                                err: err,
                                snapshot: path.snapshot,
                                errMsg: err != null ? err.message : undefined,
                                rows
                            });
                        }
                    });
                    db.close();
                } else {
                    try{
                        const end_execute_time = new Date();
                        var diff = (end_execute_time-begin_execute_time);
                        configLog.processTime=diff;
                        configLog.errMsg = path.err;
                        module_core_recordRequestLog(system, system.user.email, 'sqlite', configLog);
                    }catch(e){}
                    finally{
                        callback({
                            err: null,
                            snapshot: path.snapshot
                        }, 302);
                    }
                }
            } else {
                let db = new sqlite3.Database(path.path, sqlite3.OPEN_READONLY);
                db.all(module_core_selectOnlySql(data.sql), params(system.user, data.params), (err, rows) => {
                    try{
                        const end_execute_time = new Date();
                        var diff = (end_execute_time-begin_execute_time);
                        configLog.processTime=diff;
                        configLog.errMsg = path.err;
                        module_core_recordRequestLog(system, system.user.email, 'sqlite', configLog);
                    }catch(e){}
                    finally{
                        callback({
                            err: err,
                            snapshot: path.snapshot,
                            errMsg: err != null ? err.message : undefined,
                            rows
                        });
                    }
                });
                db.close();
            }
        }
    });


}

module_sqlite_get = function (system, data, callback) {
    const begin_execute_time = new Date();
    let configLog={apiName:'module_sqlite_get', config: JSON.stringify(data)};

    module_sqlite_path(system, data, function (path) {
        if (path.err) {
            try{
                const end_execute_time = new Date();
                var diff = (end_execute_time-begin_execute_time);
                configLog.processTime=diff;
                configLog.errMsg = path.err;
                module_core_recordRequestLog(system, system.user.email, 'sqlite', configLog);
            }catch(e){}
            finally{
                callback({
                    err: path.err,
                    snapshot: path.snapshot,
                    errMsg: path.err,
                    row: null
                });
            }
        } else {
            let db = new sqlite3.Database(path.path, sqlite3.OPEN_READONLY);
            db.get(module_core_selectOnlySql(data.sql), params(system.user, data.params), (err, row) => {
                try{
                    const end_execute_time = new Date();
                    var diff = (end_execute_time-begin_execute_time);
                    configLog.processTime=diff;
                    configLog.errMsg = path.err;
                    module_core_recordRequestLog(system, system.user.email, 'sqlite', configLog);
                }catch(e){}
                finally{
                    callback({
                        err: err,
                        snapshot: path.snapshot,
                        errMsg: err != null ? err.message : undefined,
                        row
                    });
                }
            });
            db.close();
        }
    });
}

module_sqlite_path = function (system, data, callback) {
    let snapshot=''
    let snapshot_path=''
    if(data&&data.snapshot){ 
        snapshot=data.snapshot
        snapshot_path=`/snapshot/${snapshot}` 
    }
    if (data.dbName) {
        let db = new sqlite3.Database(system.sysPath + "/db/system.db", sqlite3.OPEN_READONLY);
        db.get("SELECT dbId from tbl_db WHERE dbName = ? ", [data.dbName], (err, row) => {
            if (row && fs.existsSync(system.sysPath + "/db/database_user/" + row.dbId + ".db")) {
                callback({
                    err: null,
                    path: system.sysPath + "/db/database_user/" + row.dbId + ".db"
                });
            } else {
                callback({
                    err: "Database not found"
                })
            }
        });
        db.close();
    } else {
        callback({
            err: null,
            snapshot,
            path: system.sysPath +snapshot_path + (data.db == undefined ? "/db/database.db" : "/db/" + data.db + ".db")
        })
    }
}

function module_sqlite_formData_received(system, data, callback) {
    module_core_checkPermission(system, 'c', 'database', (result) => {
        if (result) {
            let db = new sqlite3.Database(system.sysPath + "/db/system.db");
            db.parallelize(() => {
                db.get("SELECT tbl, tblId FROM tbl007 WHERE hash = ?", [data.fields["$_amgCnt_formData_submit_tbl_$"]], (err, row) => {
                    if (err != null) {
                        callback({
                            err: err,
                            errMsg: err.message
                        });
                    } else if (row == undefined) {
                        callback({
                            err: "TABLE NOT FOUND",
                            errMsg: "Table not found"
                        });
                    } else {
                        db.all("SELECT hash, col, colId FROM tbl007 WHERE tbl = ?", [row.tbl], (err, rows) => {
                            if (err != null) {
                                callback({
                                    err: err,
                                    errMsg: err.message
                                });
                            } else {
                                let col = [],
                                    params = [],
                                    map = {},
                                    file = [];
                                delete data.fields["$_amgCnt_formData_submit_tbl_$"];
                                delete data.fields["$_amgCnt_formData_submit_token_$"];
                                rows.forEach(function (d) {
                                    map[d.hash] = {
                                        col: d.col,
                                        colId: d.colId
                                    };
                                });
                                for (let name in data.fields) {
                                    col.push(map[name].col);
                                    params.push(data.fields[name]);
                                }
                                for (let name in data.files) {
                                    let f = [];
                                    if (data.files[name].length != undefined) {
                                        for (let i = 0; i < data.files[name].length; i++) {
                                            let hash = module_core_hash(data.files[name][i].name + module_core_genId(16));
                                            f.push(data.files[name][i].name + "/" + hash);
                                            file.push({
                                                hash: hash,
                                                path: data.files[name][i].path,
                                                sql: "INSERT INTO files (id, name, type, size, lastModifiedDate) VALUES(?,?,?,?,?)",
                                                params: [hash, data.files[name][i].name, data.files[name][i].type, data.files[name][i].size, data.files[name][i].lastModifiedDate]
                                            });
                                        }
                                    } else {
                                        let hash = module_core_hash(data.files[name].name + module_core_genId(16));
                                        f.push(data.files[name].name + "/" + hash);
                                        file.push({
                                            hash: hash,
                                            path: data.files[name].path,
                                            sql: "INSERT INTO files (id, name, type, size, lastModifiedDate) VALUES(?,?,?,?,?)",
                                            params: [hash, data.files[name].name, data.files[name].type, data.files[name].size, data.files[name].lastModifiedDate]
                                        });
                                    }
                                    col.push(map[name].col);
                                    params.push(f.join("|"));
                                }
                                if (col.length > 0) {
                                    let db = new sqlite3.Database(system.sysPath + "/db/database.db");
                                    db.parallelize(() => {
                                        db.run("INSERT INTO " + row.tbl + "(" + col.join(",") + ") VALUES(" + col.map(d => '?').join(",") + ");", params, (err) => {
                                            if (err != null) {
                                                file.forEach(function (d) {
                                                    fs.unlink(d.path, () => {});
                                                });
                                                callback({
                                                    err: err,
                                                    errMsg: err.message
                                                });
                                            } else {
                                                let db2 = new sqlite3.Database(system.sysPath + "/db/files.db");
                                                db2.parallelize(() => {
                                                    file.forEach(function (d) {
                                                        db2.run(d.sql, d.params);
                                                        fs.renameSync(d.path, system.sysPath + "/drive/root/" + d.hash);
                                                    });
                                                });
                                                db2.close();
                                                callback({
                                                    err: err,
                                                    errMsg: err != null ? err.message : undefined
                                                });
                                            }
                                        });
                                    });
                                    db.close();
                                } else {
                                    callback({
                                        err: "NO COLUMN TO INSERT",
                                        errMsg: "NO COLUMN TO INSERT"
                                    });
                                }
                            }
                        });
                    }
                });
            });
            db.close();
        } else {
            callback({
                err: "No permission to insert row",
                errMsg: "No permission to insert row"
            });
        }
    });
}

function module_sqlite_hash(structure) {
    structure.hash = module_core_hash(structure.name);
    structure.column.forEach(function (d) {
        d.hash = module_core_hash(structure.name + d.name) + module_core_genId(8);
    });
}
//++ data dictionary
function module_sqlite_updateDataDictionaryFormat(system, data, callback) {
    let db = new sqlite3.Database(system.sysPath + "/db/system.db");
    switch (data.mode) {
        case 0:
            /*create*/
            db.parallelize(() => {
                db.run("INSERT INTO tbl002(tblId, structure, type, user_created, time_created, user_updated, time_updated) VALUES(?,?,?,?,?,?,?)", [data.tblId, JSON.stringify(data.structure), data.type, data.user_created, data.time_created, data.user_updated, data.time_updated]);
            });
            break;
        case 1:
            /*update*/
            //db.serialize(()=>{
            db.parallelize(() => {
                db.run("UPDATE tbl002 SET structure = ?, type = ?, user_updated = ?, time_updated=? WHERE tblId = ?", [JSON.stringify(data.structure), data.type, data.user_updated, data.time_updated, data.tblId]);
            });
            //});
            break;
        case 2:
            /*delete*/
            db.parallelize(() => {
                db.run("DELETE FROM tbl002 WHERE tblId IN (" + data.tblId.map(d => ("?")) + ")", data.tblId);
            });
            break;
    }
    db.close();
    callback({
        err: {
            sql1: data.mode,
            sql2: "INSERT INTO tbl002(tblId, structure, type, user_created, time_created, user_updated, time_updated) VALUES(?,?,?,?,?,?,?)",
            sql3: [data.tblId, JSON.stringify(data.structure), data.type, data.user_created, data.time_created, data.user_updated, data.time_updated]
        },
        errMsg: undefined
    });
}

function module_sqlite_updateStructure(system, data) {
    let db = new sqlite3.Database(system.sysPath + "/db/system.db");
    let sql, params;
    switch (data.mode) {
        case 0:
            /*create*/
            module_sqlite_hash(data.structure);
            sql = "INSERT INTO tbl007(hash, tbl, col, tblId, colId, dbId) VALUES(?,?,?,?,?,?)";
            params = [data.structure.hash, data.structure.name, null, data.structure.id, null, data.dbId];
            data.structure.column.forEach(function (d) {
                sql += ", (?,?,?,?,?,?)";
                params = params.concat([d.hash, data.structure.name, d.name, data.structure.id, d.id, data.dbId]);
            });
            db.parallelize(() => {
                db.run(sql, params);
                db.run("INSERT INTO tbl002(tblId, structure, type, user_created, time_created, user_updated, time_updated, dbId) VALUES(?,?,?,?,?,?,?,?)", [data.tblId, JSON.stringify(data.structure), data.type, data.user_created, data.time_created, data.user_updated, data.time_updated, data.dbId]);
            });
            break;
        case 1:
            /*update*/
            module_sqlite_hash(data.structure);
            sql = "INSERT INTO tbl007(hash, tbl, col, tblId, colId) VALUES(?,?,?,?,?)";
            params = [data.structure.hash, data.structure.name, null, data.structure.id, null];
            data.structure.column.forEach(function (d) {
                sql += ", (?,?,?,?,?)";
                params = params.concat([d.hash, data.structure.name, d.name, data.structure.id, d.id]);
            });
            db.serialize(() => {
                db.run("DELETE FROM tbl007 WHERE tbl = ?", [data.name0]);
                db.parallelize(() => {
                    db.run(sql, params);
                    db.run("UPDATE tbl002 SET structure = ?, type = ?, user_updated = ?, time_updated=? WHERE tblId = ?", [JSON.stringify(data.structure), data.type, data.user_updated, data.time_updated, data.tblId]);
                });
            });
            break;
        case 2:
            /*delete*/
            db.parallelize(() => {
                db.run("DELETE FROM tbl002 WHERE tblId IN (" + data.tblId.map(d => ("?")) + ")", data.tblId);
                db.run("DELETE FROM tbl007 WHERE tbl IN (" + data.tblId.map(d => ("?")) + ")", data.name);
                //db.run("UPDATE _amgcnt_uploaded_files SET tbl = NULL WHERE tbl IN (" + data.tblId.map(d=>("?")) +")", data.name);
            });

            break;
    }
    db.close();
}

function module_sqlite_updateTable(system, data, callback) {
    let count = 0;
    let errArr = [];

    function success() {
        if (errArr.length == 0) {
            callback({
                err: null,
                errMsg: undefined
            });
        } else {
            callback({
                err: errArr.map(d => d.err).join("\r\n"),
                errMsg: errArr.map(d => d.errMsg).join("\r\n")
            });
        }
    }
    data.forEach(function (d) {
        let db = new sqlite3.Database(system.sysPath + (d.dbId == undefined ? "/db/database.db" : "/db/database_user/" + d.dbId + ".db"));
        db.parallelize(() => {
            db.exec("PRAGMA foreign_keys=off;BEGIN TRANSACTION;" + d.sql, (err) => {
                if (err == null) {
                    db.exec("COMMIT;PRAGMA foreign_keys=on;");
                    module_sqlite_updateStructure(system, d);
                } else {
                    errArr.push({
                        err: err,
                        errMsg: err != null ? err.message : undefined
                    })
                    db.exec("ROLLBACK;PRAGMA foreign_keys=on;");
                }
                if (++count == data.length) {
                    success();
                }
            });
        });
        db.close();
    })
}

function module_sqlite_read(system, data, callback) {
    module_core_checkPermission(system, 'r', 'database', (result) => {
        if (result) {
            if (data.mode == 0) {
                module_sqlite_get(system, {
                    sql: data.sql,
                    params: data.params,
                    dbName: data.dbName,
                    snapshot: data.snapshot
                }, callback);
            } else {
                module_sqlite_all(system, {
                    sql: data.sql,
                    params: data.params,
                    dbName: data.dbName,
                    snapshot: data.snapshot
                }, callback);
            }
        } else {
            callback({
                err: "No permission to read database",
                errMsg: "No permission to read database"
            });
        }
    });
}

function module_sqlite_create(system, data, callback) {
    module_core_checkPermission(system, 'c', 'database', (result) => {
        if (result) {
            module_sqlite_run(system, {
                sql: data.sql.search(/(\s*update\s+.+\s+set\s+)|(\s*delete\s+.+\s+from\s+)/gi) == -1 ? data.sql : '',
                params: data.params,
                dbName: data.dbName,
                snapshot: data.snapshot
            }, callback);
        } else {
            callback({
                err: "No permission to create rows in database",
                errMsg: "No permission to create rows in database"
            });
        }
    });
}

function module_sqlite_update(system, data, callback) {
    module_core_checkPermission(system, 'u', 'database', (result) => {
        if (result) {
            module_sqlite_run(system, {
                sql: data.sql.search(/(\s*insert\s+.+\s+into\s+)|(\s*delete\s+.+\s+from\s+)/gi) == -1 ? data.sql : '',
                params: data.params,
                dbName: data.dbName,
                snapshot: data.snapshot
            }, callback);
        } else {
            callback({
                err: "No permission to update rows in database",
                errMsg: "No permission to update rows in database"
            });
        }
    });
}

function module_sqlite_delete(system, data, callback) {
    module_core_checkPermission(system, 'd', 'database', (result) => {
        if (result) {
            module_sqlite_run(system, {
                sql: data.sql.search(/(\s*insert\s+.+\s+into\s+)|(\s*update\s+.+\s+set\s+)/gi) == -1 ? data.sql : '',
                params: data.params,
                dbName: data.dbName,
                snapshot: data.snapshot
            }, callback);
        } else {
            callback({
                err: "No permission to delete rows in database",
                errMsg: "No permission to delete rows in database"
            });
        }
    });
}

function module_sqlite_insertAPI(system, data, callback) {
    let params = [],
        values = [];
    try {
        data.rows.forEach(function (d) {
            params = params.concat(data.column.map(d1 => d[d1]));
            values = values.concat('(' + data.column.map(d1 => '?') + ')');
        });
        let sql = "INSERT INTO \"" + data.table + "\"(" + data.column.map(d => ('"' + d + '"')) + ") VALUES" + values + ";";
        let db = new sqlite3.Database(system.sysPath + "/db/database.db");
        db.run(sql, params, (err) => {
            callback({
                err: err,
                errMsg: err != null ? err.message : undefined
            });
        });
        db.close();
    } catch (e) {
        callback({
            err: "Incorrect format"
        });
    }
}

function module_sqlite_updateDatabase(system, data, callback) {
    let db = new sqlite3.Database(system.sysPath + "/db/system.db");
    switch (data.mode) {
        case 0:
            /*New database*/
            db.run("INSERT INTO tbl_db(dbId, dbName, description, user_created, user_updated, time_created, time_updated) VALUES(?,?,?,?,?,?,?)", [data.dbData.data.id, data.dbData.data.name, data.dbData.data.description, data.dbData.user_created, data.dbData.user_updated, data.dbData.time_created, data.dbData.time_updated], (err) => {
                if (err) {
                    callback({
                        symbol: false,
                        text: "Error Found: " + err.code
                    })
                } else {
                    copyDb();
                }
            });

            function copyDb() {
                fs.copyFile("./core/template/db_template.db", system.sysPath + "/db/database_user/" + data.dbData.data.id + ".db", (err) => {
                    if (err) {
                        console.log("Error Found:", err);
                    } else {
                        let c = {
                            symbol: true,
                            text: "Database: \" " + data.dbData.data.name + " \" has been created."
                        };
                        callback(c);
                    }
                });
            }
            break;
        case 1:
            /*Edit database info*/
            db.run("UPDATE tbl_db SET dbName = ?, user_updated = ?, description = ?, time_updated = ? WHERE dbId = ?;", [data.dbData.data.name, data.dbData.user_updated, data.dbData.data.description, data.dbData.time_updated, data.dbData.data.id], (err) => {
                if (err) {
                    callback({
                        symbol: false,
                        text: "Error Found: " + err.code
                    })
                }
            });

            break;
        case 2:
            /*Delete database*/
            data.dbData.dbId.forEach(dbId => {
                db.parallelize(() => {
                    db.run("DELETE FROM tbl_db WHERE dbId = ?", dbId, (err) => {
                        if (err) {
                            callback({
                                symbol: false,
                                text: "Error Found: " + err.code
                            })
                        } else {
                            unlinkDb()
                        }
                    });
                    db.run("DELETE FROM tbl002 WHERE dbId = ?", dbId);
                    db.run("DELETE FROM tbl007 WHERE dbId = ?", dbId);
                });

                function unlinkDb() {
                    fs.unlink(system.sysPath + "/db/database_user/" + dbId + ".db", function (err) {
                        if (err) {
                            console.log("Error Found:", err);
                            callback({
                                symbol: false,
                                text: "Error Found: " + err.code
                            })
                        } else {
                            let c = {
                                symbol: true,
                                text: "Database has been deleted."
                            };
                            callback(c);
                        }
                    });
                }
            });
            break;
    }
    db.close();
}